@@ -294,7 +294,21 @@ var SyscallsLibrary = {
     var old = SYSCALLS.getStreamFromFD();
     return FS.open(old.path, old.flags, 0).fd;
   },
-  __syscall42: '__syscall51',      // pipe
+  __syscall42__deps: ['$PIPEFS'],
+  __syscall42: function(which, varargs) { // pipe
+    var fdPtr = SYSCALLS.get();
+
+    if (fdPtr == 0) {
+      throw new FS.ErrnoError(ERRNO_CODES.EFAULT);
+    }
+
+    var res = PIPEFS.createPipe();
+
+    {{{ makeSetValue('fdPtr', 0, 'res.readable_fd', 'i32') }}};
+    {{{ makeSetValue('fdPtr', 4, 'res.writable_fd', 'i32') }}};
+
+    return 0;
+  },
   __syscall51: function(which, varargs) { // acct
     return -ERRNO_CODES.ENOSYS; // unsupported features
   },
@@ -332,6 +346,12 @@ var SyscallsLibrary = {
         var argp = SYSCALLS.get();
         return FS.ioctl(stream, op, argp);
       }
+      case {{{ cDefine('TIOCGWINSZ') }}}: {
+        // TODO: in theory we should write to the winsize struct that gets
+        // passed in, but for now musl doesn't read anything on it
+        if (!stream.tty) return -ERRNO_CODES.ENOTTY;
+        return 0;
+      }
       default: abort('bad ioctl syscall ' + op);
     }
 #endif // NO_FILESYSTEM
@@ -659,8 +679,8 @@ var SyscallsLibrary = {
   },
   __syscall140: function(which, varargs) { // llseek
     var stream = SYSCALLS.getStreamFromFD(), offset_high = SYSCALLS.get(), offset_low = SYSCALLS.get(), result = SYSCALLS.get(), whence = SYSCALLS.get();
+    // NOTE: offset_high is unused - Emscripten's off_t is 32-bit
     var offset = offset_low;
-    assert(offset_high === 0);
     FS.llseek(stream, offset, whence);
     {{{ makeSetValue('result', '0', 'stream.position', 'i32') }}};
     if (stream.getdents && offset === 0 && whence === {{{ cDefine('SEEK_SET') }}}) stream.getdents = null; // reset readdir state
@@ -833,23 +853,27 @@ var SyscallsLibrary = {
     }
     return nonzero;
   },
+  __syscall178: function(which, varargs) { // rt_sigqueueinfo
+#if SYSCALL_DEBUG
+    Module.printErr('warning: ignoring SYS_rt_sigqueueinfo');
+#endif
+    return 0;
+  },
   __syscall180: function(which, varargs) { // pread64
     var stream = SYSCALLS.getStreamFromFD(), buf = SYSCALLS.get(), count = SYSCALLS.get(), zero = SYSCALLS.getZero(), offset = SYSCALLS.get64();
     return FS.read(stream, {{{ heapAndOffset('HEAP8', 'buf') }}}, count, offset);
   },
   __syscall181: function(which, varargs) { // pwrite64
-#if SYSCALL_DEBUG
-    Module.printErr('warning: untested syscall');
-#endif
     var stream = SYSCALLS.getStreamFromFD(), buf = SYSCALLS.get(), count = SYSCALLS.get(), zero = SYSCALLS.getZero(), offset = SYSCALLS.get64();
-    return FS.write(stream, {{{ heapAndOffset('HEAP8', 'buf') }}}, nbyte, offset);
+    return FS.write(stream, {{{ heapAndOffset('HEAP8', 'buf') }}}, count, offset);
   },
   __syscall183: function(which, varargs) { // getcwd
     var buf = SYSCALLS.get(), size = SYSCALLS.get();
     if (size === 0) return -ERRNO_CODES.EINVAL;
     var cwd = FS.cwd();
-    if (size < cwd.length + 1) return -ERRNO_CODES.ERANGE;
-    writeAsciiToMemory(cwd, buf);
+    var cwdLengthInBytes = lengthBytesUTF8(cwd);
+    if (size < cwdLengthInBytes + 1) return -ERRNO_CODES.ERANGE;
+    stringToUTF8(cwd, buf, size);
     return buf;
   },
   __syscall191: function(which, varargs) { // ugetrlimit
@@ -869,7 +893,7 @@ var SyscallsLibrary = {
     var ptr;
     var allocated = false;
     if (fd === -1) {
-      ptr = _malloc(len);
+      ptr = _memalign(PAGE_SIZE, len);
       if (!ptr) return -ERRNO_CODES.ENOMEM;
       _memset(ptr, 0, len);
       allocated = true;
@@ -969,11 +993,10 @@ var SyscallsLibrary = {
       stream.getdents = FS.readdir(stream.path);
     }
     var pos = 0;
-    while (stream.getdents.length > 0 && pos + {{{ C_STRUCTS.dirent.__size__ }}} < count) {
+    while (stream.getdents.length > 0 && pos + {{{ C_STRUCTS.dirent.__size__ }}} <= count) {
       var id;
       var type;
       var name = stream.getdents.pop();
-      assert(name.length < 256); // limit of dirent struct
       if (name[0] === '.') {
         id = 1;
         type = 4; // DT_DIR
@@ -989,16 +1012,19 @@ var SyscallsLibrary = {
       {{{ makeSetValue('dirp + pos', C_STRUCTS.dirent.d_off, 'stream.position', 'i32') }}};
       {{{ makeSetValue('dirp + pos', C_STRUCTS.dirent.d_reclen, C_STRUCTS.dirent.__size__, 'i16') }}};
       {{{ makeSetValue('dirp + pos', C_STRUCTS.dirent.d_type, 'type', 'i8') }}};
-      for (var i = 0; i < name.length; i++) {
-        {{{ makeSetValue('dirp + pos', C_STRUCTS.dirent.d_name + ' + i', 'name.charCodeAt(i)', 'i8') }}};
-      }
-      {{{ makeSetValue('dirp + pos', C_STRUCTS.dirent.d_name + ' + i', '0', 'i8') }}};
+      stringToUTF8(name, dirp + pos + {{{ C_STRUCTS.dirent.d_name }}}, 256);
       pos += {{{ C_STRUCTS.dirent.__size__ }}};
     }
     return pos;
   },
   __syscall221__deps: ['__setErrNo'],
   __syscall221: function(which, varargs) { // fcntl64
+#if NO_FILESYSTEM
+#if SYSCALL_DEBUG
+    Module.printErr('no-op in fcntl64 syscall due to NO_FILESYSTEM');
+#endif
+    return 0;
+#else
     var stream = SYSCALLS.getStreamFromFD(), cmd = SYSCALLS.get();
     switch (cmd) {
       case {{{ cDefine('F_DUPFD') }}}: {
@@ -1047,6 +1073,7 @@ var SyscallsLibrary = {
         return -ERRNO_CODES.EINVAL;
       }
     }
+#endif // NO_FILESYSTEM
   },
   __syscall265: function(which, varargs) { // clock_nanosleep
 #if SYSCALL_DEBUG
@@ -1234,6 +1261,12 @@ var SyscallsLibrary = {
     var stream = SYSCALLS.getStreamFromFD(), iov = SYSCALLS.get(), iovcnt = SYSCALLS.get(), offset = SYSCALLS.get();
     return SYSCALLS.doWritev(stream, iov, iovcnt, offset);
   },
+  __syscall337: function(which, varargs) { // recvmmsg
+#if SYSCALL_DEBUG
+    Module.printErr('warning: ignoring SYS_recvmmsg');
+#endif
+    return 0;
+  },
   __syscall340: function(which, varargs) { // prlimit64
     var pid = SYSCALLS.get(), resource = SYSCALLS.get(), new_limit = SYSCALLS.get(), old_limit = SYSCALLS.get();
     if (old_limit) { // just report no limits
@@ -1244,6 +1277,12 @@ var SyscallsLibrary = {
     }
     return 0;
   },
+  __syscall345: function(which, varargs) { // sendmmsg
+#if SYSCALL_DEBUG
+    Module.printErr('warning: ignoring SYS_sendmmsg');
+#endif
+    return 0;
+  },
 };
 
 if (SYSCALL_DEBUG) {
@@ -1580,6 +1619,7 @@ if (SYSCALL_DEBUG) {
     SYS_inotify_init1: 332,
     SYS_preadv: 333,
     SYS_pwritev: 334,
+    SYS_recvmmsg: 337,
     SYS_prlimit64: 340,
     SYS_name_to_handle_at: 341,
     SYS_open_by_handle_at: 342,
@@ -1667,4 +1707,3 @@ SyscallsLibrary.emscripten_syscall = eval('(' + switcher + ')');
 #endif
 
 mergeInto(LibraryManager.library, SyscallsLibrary);
-
